/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch das
 * Ministerium für Energiewende, Klimaschutz, Umwelt und Natur
 * Zentrales IT-Management
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.formcycle;

import static java.util.Objects.*;
import static org.apache.commons.lang3.StringUtils.*;

import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.util.Collection;
import java.util.Optional;

import org.apache.commons.io.IOUtils;
import org.apache.http.HttpHost;
import org.apache.http.HttpStatus;
import org.apache.http.auth.AuthScope;
import org.apache.http.auth.UsernamePasswordCredentials;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.protocol.HttpClientContext;
import org.apache.http.conn.routing.HttpRoutePlanner;
import org.apache.http.impl.client.BasicAuthCache;
import org.apache.http.impl.client.BasicCredentialsProvider;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.impl.conn.DefaultProxyRoutePlanner;
import org.apache.http.protocol.HttpContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.ozgcloud.eingang.formcycle.FormCycleConfirmationResponse;
import de.ozgcloud.eingang.formcycle.FormCycleFormData;
import de.ozgcloud.formcycle.attachment.FormcycleAttachment;
import lombok.RequiredArgsConstructor;

@RequiredArgsConstructor
public class OzgCloudFormDataHttpClient {

	static final int HTTPCLIENT_TIMEOUT = 10 * 60 * 1000;

	private static final Logger LOG = LoggerFactory.getLogger(OzgCloudFormDataHttpClient.class);

	private final OzgCloudConfig config;
	private final SystemPropertiesProvider systemPropertiesProvider;
	private Optional<ProxyConfig> systemProxyConfig;

	private BasicCredentialsProvider credentialsProvider;
	private HttpClientContext context;
	private HttpHost proxyHost;

	public FormCycleConfirmationResponse send(FormCycleFormData formData, Collection<FormcycleAttachment> attachments,
			Collection<FormcycleAttachment> representations) throws IOException {

		try (
				var closeableHttpClient = createCloseableClient();
				var closeableResponse = closeableHttpClient.execute(buildPostRequest(formData, attachments, representations), getContext())) {
			return parseResponse(closeableResponse);
		} catch (Exception e) {
			LOG.info("Used OzgCloudConfig: {}", config);
			throw e;
		}
	}

	CloseableHttpClient createCloseableClient() {
		var clientBuilder = HttpClientBuilder.create().setDefaultRequestConfig(createHttpClientConfig());
		addProxy(clientBuilder);
		return clientBuilder.build();
	}

	void addProxy(HttpClientBuilder httpClientBuilder) {
		Optional.ofNullable(getProxyConfig()).ifPresent(
				host -> configureProxy(httpClientBuilder));
	}

	void configureProxy(HttpClientBuilder httpClientBuilder) {
		httpClientBuilder.setRoutePlanner(createHttpRouterPlanner());
		httpClientBuilder.setDefaultCredentialsProvider(getCredentialsProvider());
	}

	ProxyConfig getProxyConfig() {
		var proxyConfig = config.getProxyConfig();
		if (isNoneBlank(proxyConfig.getHost())) {
			return proxyConfig;
		}
		if (isNull(systemProxyConfig)) {
			systemProxyConfig = loadSystemProxyConfig();
		}
		return systemProxyConfig.orElse(null);
	}

	Optional<ProxyConfig> loadSystemProxyConfig() {
		ProxyConfig value = loadHttpsProxyConfig();
		ProxyConfig sysProxyConfig = Optional.ofNullable(value).orElseGet(this::loadHttpProxyConfig);
		return Optional.ofNullable(sysProxyConfig);
	}

	ProxyConfig loadHttpsProxyConfig() {
		var systemProxyHost = systemPropertiesProvider.getHttpsProxyHost();
		if (isBlank(systemProxyHost)) {
			return null;
		}
		return ProxyConfig.builder()
				.host(systemProxyHost)
				.port(Utils.parsePort(systemPropertiesProvider.getHttpsProxyPort()))
				.user(systemPropertiesProvider.getHttpsProxyUser())
				.password(systemPropertiesProvider.getHttpsProxyPassword())
				.build();
	}

	ProxyConfig loadHttpProxyConfig() {
		var systemProxyHost = systemPropertiesProvider.getHttpProxyHost();
		if (isBlank(systemProxyHost)) {
			return null;
		}
		return ProxyConfig.builder()
				.host(systemProxyHost)
				.port(Utils.parsePort(systemPropertiesProvider.getHttpProxyPort()))
				.user(systemPropertiesProvider.getHttpProxyUser())
				.password(systemPropertiesProvider.getHttpProxyPassword())
				.build();
	}

	HttpHost getProxyHost() {
		if (isNull(proxyHost)) {
			proxyHost = createProxyHost();
		}
		return proxyHost;
	}

	HttpHost createProxyHost() {
		var proxyConfig = getProxyConfig();
		return new HttpHost(proxyConfig.getHost(), proxyConfig.getPort());
	}

	BasicCredentialsProvider getCredentialsProvider() {
		if (isNull(credentialsProvider)) {
			credentialsProvider = createCredentialsProvider();
		}
		return credentialsProvider;
	}

	BasicCredentialsProvider createCredentialsProvider() {
		var basicCredentialsProvider = new BasicCredentialsProvider();
		var proxyConfig = getProxyConfig();
		basicCredentialsProvider.setCredentials(new AuthScope(getProxyHost()),
				new UsernamePasswordCredentials(proxyConfig.getUser(), proxyConfig.getPassword()));
		return basicCredentialsProvider;
	}

	HttpRoutePlanner createHttpRouterPlanner() {
		return new DefaultProxyRoutePlanner(getProxyHost());
	}

	RequestConfig createHttpClientConfig() {
		return RequestConfig.custom()
				.setConnectTimeout(HTTPCLIENT_TIMEOUT)
				.setConnectionRequestTimeout(HTTPCLIENT_TIMEOUT)
				.setSocketTimeout(HTTPCLIENT_TIMEOUT).build();
	}

	HttpPost buildPostRequest(FormCycleFormData formData, Collection<FormcycleAttachment> attachments,
			Collection<FormcycleAttachment> representations) {
		var builder = new HttpPostRequestBuilder(config).withFormData(formData);
		Optional.ofNullable(attachments).ifPresent(builder::withAttachments);
		Optional.ofNullable(representations).ifPresent(builder::withRepresentations);
		return builder.build();
	}

	HttpContext getContext() {
		if (isNull(context)) {
			context = createContext();
		}
		return context;
	}

	HttpClientContext createContext() {
		var clientContext = HttpClientContext.create();
		if (nonNull(getProxyConfig())) {
			clientContext.setCredentialsProvider(getCredentialsProvider());
			clientContext.setAuthCache(new BasicAuthCache());
		}
		return clientContext;
	}

	FormCycleConfirmationResponse parseResponse(CloseableHttpResponse response) throws IOException {
		checkResponseStatus(response);
		InputStream responseContent = response.getEntity().getContent();
		return FormCycleConfirmationResponse.parseFrom(responseContent);
	}

	private void checkResponseStatus(CloseableHttpResponse response) {
		var statusLine = response.getStatusLine();
		if (statusLine.getStatusCode() != HttpStatus.SC_OK) {
			throw new UnexpectedStatusCodeException(response);
		}
	}

	static class UnexpectedStatusCodeException extends RuntimeException {
		private static final String MESSAGE = "Unexcepted Status received: %s";

		UnexpectedStatusCodeException(CloseableHttpResponse response) {
			super(buildFullMessage(response));
		}

		private static String buildFullMessage(CloseableHttpResponse response) {
			try {
				StringBuilder messageBuilder = new StringBuilder(String.format(MESSAGE, response.getStatusLine())).append("\n");

				IOUtils.readLines(response.getEntity().getContent(), StandardCharsets.UTF_8)
						.stream().forEach(line -> messageBuilder.append(line).append("\n"));

				return messageBuilder.toString();
			} catch (IOException e) {
				return MESSAGE + "\nError reading body";
			}
		}
	}
}
