/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch das
 * Ministerium für Energiewende, Klimaschutz, Umwelt und Natur
 * Zentrales IT-Management
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.formcycle;

import static java.util.Objects.*;

import java.io.IOException;
import java.util.EnumMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.Supplier;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.ozgcloud.formcycle.attachment.AttachmentMapper;
import de.ozgcloud.formcycle.attachment.AttachmentType;
import de.ozgcloud.formcycle.attachment.FormcycleAttachment;
import de.ozgcloud.formcycle.errorhandling.NodeThrewExceptionFactory;
import de.ozgcloud.formcycle.formdata.OzgCloudFormDataMapper;
import de.ozgcloud.formcycle.formdata.PluginFormDataAdapter;
import de.xima.fc.entities.Attachment;
import de.xima.fc.exceptions.AbstractAbruptCompletionException;
import de.xima.fc.exceptions.NodeThrewException;
import lombok.RequiredArgsConstructor;

@RequiredArgsConstructor
public final class WorkflowElementNodeExecutor {

	private static final Logger LOG = LoggerFactory.getLogger(WorkflowElementNodeExecutor.class);

	private final OzgCloudFormDataMapper ozgCloudFormDataMapper;
	private final OzgCloudFormDataHttpClient ozgHttpClient;
	private final NodeThrewExceptionFactory exceptionFactory;
	private final AttachmentMapper attachmentMapper;
	private final PluginFormDataAdapter pluginFormDataAdapter;
	private final Supplier<List<Attachment>> attachmentsSupplier;

	public ExecutionResult execute() throws AbstractAbruptCompletionException {
		LOG.debug("Executing plugin WorkflowElementNodePlugin");
		var formData = pluginFormDataAdapter.readFormData();
		var formcycleFormData = ozgCloudFormDataMapper.map(formData);
		var attachments = getAttachedFiles(formData.getAttachmentUuids());

		try {
			var formCycleConfirmationResponse = ozgHttpClient.send(formcycleFormData, attachments.get(AttachmentType.ATTACHMENT),
					attachments.get(AttachmentType.REPRESENTATION));
			LOG.debug("Formcycle adapter response: {}", formCycleConfirmationResponse);

			return ExecutionResult.builder().vorgangnummer(formCycleConfirmationResponse.getVorgangNummer()).build();

		} catch (IOException | RuntimeException e) {
			throw exceptionFactory.createCannotSendException("Cannot send data to formcycle adapter in OZG cloud.", e);
		}
	}

	Map<AttachmentType, Set<FormcycleAttachment>> getAttachedFiles(Set<String> uploadFieldUuids) throws NodeThrewException {
		var resultMap = new EnumMap<AttachmentType, Set<FormcycleAttachment>>(AttachmentType.class);
		for (Attachment attachment : attachmentsSupplier.get()) {
			if (uploadFieldUuids.contains(attachment.getUUID())) {
				resultMap.computeIfAbsent(AttachmentType.ATTACHMENT, e -> new HashSet<>()).add(attachmentMapper.map(attachment));
				continue;
			}
			resultMap.computeIfAbsent(AttachmentType.REPRESENTATION, e -> new HashSet<>()).add(attachmentMapper.map(attachment));
		}
		if (isNull(resultMap.get(AttachmentType.REPRESENTATION))) {
			LOG.warn("Representation is missing. Activate checkbox 'An den Vorgang anhaengen' in 'Word-Datei befuellen' plugin");
		}
		return resultMap;
	}

}
