/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch das
 * Ministerium für Energiewende, Klimaschutz, Umwelt und Natur
 * Zentrales IT-Management
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.formcycle.formdata;

import static de.ozgcloud.formcycle.formdata.FormDataTestFactory.*;
import static de.ozgcloud.formcycle.formdata.OzgCloudFormDataMapper.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.List;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.Spy;

import de.ozgcloud.eingang.formcycle.FormCycleAttachmentGroup;
import de.ozgcloud.eingang.formcycle.FormCycleFormHeader;
import de.ozgcloud.eingang.formcycle.FormCycleServiceKonto;
import de.ozgcloud.vorgang.vorgang.GrpcFormData;
import de.ozgcloud.vorgang.vorgang.GrpcFormField;
import de.ozgcloud.vorgang.vorgang.GrpcFormFieldOrBuilder;
import de.ozgcloud.vorgang.vorgang.GrpcSubForm;

class OzgCloudFormDataMapperTest {

	@Spy
	private OzgCloudFormDataMapper mapper = new OzgCloudFormDataMapper();

	@Nested
	@DisplayName("Map FormData")
	class TestMapFormData {

		private FormNode formNode = FormNodeTestFactory.create();

		@Nested
		@DisplayName("GrpcFormData")
		class TestGrpcFormData {

			private final FormData formData = FormDataTestFactory.createBuilder().fieldData(List.of(formNode)).build();

			@Test
			void shouldAddAllFields() {
				var formFields = mapper.mapToGrpcFormData(formData).getFieldList();

				assertThat(formFields).containsOnly(GrpcFormFieldTestFactory.create());
			}

			@Test
			@DisplayName("should return empty formData if no fieldData")
			void shouldReturnEmpty() {
				var formData = FormDataTestFactory.createBuilder().fieldData(Collections.emptyList()).build();

				var formFields = mapper.mapToGrpcFormData(formData);

				assertThat(formFields.getFormList()).isEmpty();
				assertThat(formFields.getFieldList()).isEmpty();
			}

			@Test
			@DisplayName("should not call mapSubform if nested element is not a container")
			void shouldNotCallMapSubform() {
				mapper.mapToGrpcFormData(formData);

				verify(mapper, never()).mapSubForm(formNode);
			}

			@Test
			@DisplayName("should not call mapFormFields if nested element is a container")
			void shouldNotCallMapFormFields() {
				formNode.setNestedElements(List.of(FormNodeTestFactory.create()));

				mapper.mapToGrpcFormData(formData);

				verify(mapper, never()).mapFormFields(formNode);
			}

			@Test
			@DisplayName("should set nested container as subform")
			void shouldSetNestedContainer() {
				var nestedNode = FormNodeTestFactory.createBuilder().values(List.of(FIELD_VALUE_1)).build();
				formNode.setNestedElements(List.of(nestedNode));

				var subform = mapper.mapToGrpcFormData(formData);

				assertThat(subform.getFormList()).hasSize(1);
			}
		}

		@Nested
		@DisplayName("FormField")
		class TestMapFormField {

			private final FormNode formNode = FormNodeTestFactory.create();

			@Test
			void shouldMapFieldData() {
				var expectedFormField = GrpcFormFieldTestFactory.create();

				var formFields = mapper.mapFormFields(formNode);

				assertThat(formFields).containsOnly(expectedFormField);
			}

			@Test
			void shouldMapAllFields() {
				var expectedFormField1 = GrpcFormFieldTestFactory.create();
				var fieldValue2 = "fieldValue2";
				var expectedFormField2 = GrpcFormFieldTestFactory.createBuilder().setValue(fieldValue2).build();
				formNode.setValues(List.of(FIELD_VALUE_1, fieldValue2));

				var formFields = mapper.mapFormFields(formNode);

				assertThat(formFields).containsOnly(expectedFormField1, expectedFormField2);
			}

			@Test
			void shouldNotMapEmptyValues() {
				var formNode = FormNodeTestFactory.createBuilder().values(Collections.emptyList()).build();

				var formFields = mapper.mapFormFields(formNode);

				assertThat(formFields).isEmpty();
			}

			@Test
			void shouldNotMapContainer() {
				formNode.getNestedElements().add(FormNodeTestFactory.create());

				var formFields = mapper.mapFormFields(formNode);

				assertThat(formFields).isEmpty();
			}

			@Test
			void shouldSetLabel() {
				var formFields = mapper.mapFormFields(formNode);

				assertThat(formFields).first().extracting(GrpcFormField::getLabel).isEqualTo(StructureMockFactory.ITEM_TITLE);
			}

			@Test
			void shouldCleanHtmlInTitle() {
				var titleWithHtml = "<p>title</p>";
				var expectedTitle = "title";
				var formNode = FormNodeTestFactory.createBuilder().title(titleWithHtml).build();

				var formField = mapper.createGrpcFormFieldBuilder(formNode);

				assertThat(formField).extracting(GrpcFormFieldOrBuilder::getLabel).isEqualTo(expectedTitle);
			}
		}

		@Nested
		@DisplayName("GrpcSubForm")
		class TestMapSubForm {

			private final FormNode nestedNode = FormNodeTestFactory.create();

			@BeforeEach
			void setUp() {
				formNode.getNestedElements().add(nestedNode);
			}

			@Test
			void shouldSetTitle() {
				var formFields = mapper.mapSubForm(formNode);

				assertThat(formFields).isPresent().get().extracting(GrpcSubForm::getTitle).isEqualTo(StructureMockFactory.ITEM_NAME);
			}

			@Test
			void shouldAddAllFields() {
				var expectedFormField = GrpcFormFieldTestFactory.create();

				var formFields = mapper.mapSubForm(formNode).get().getFieldList();

				assertThat(formFields).containsOnly(expectedFormField);
			}

			@Test
			void shouldNotMapEmptySubform() {
				formNode.setNestedElements(Collections.emptyList());

				var formFields = mapper.mapSubForm(formNode);

				assertThat(formFields).isEmpty();
			}

			@Test
			@DisplayName("should not call mapSubform if nested element is not a container")
			void shouldNotCallMapSubform() {
				mapper.mapSubForm(formNode);

				verify(mapper, never()).mapSubForm(nestedNode);
			}

			@Test
			@DisplayName("should not call mapFormFields if nested element is a container")
			void shouldNotCallMapFormFields() {
				nestedNode.setNestedElements(List.of(FormNodeTestFactory.create()));

				mapper.mapSubForm(formNode);

				verify(mapper, never()).mapFormFields(nestedNode);
			}

			@Test
			@DisplayName("should set nested container as subform")
			void shouldSetNestedContainer() {
				var endNode = FormNodeTestFactory.create();
				nestedNode.setNestedElements(List.of(endNode));

				var subform = mapper.mapSubForm(formNode).get();

				assertThat(subform.getSubFormList()).hasSize(1);
			}

			@Test
			void shouldSetLabel() {

				var subform = mapper.mapSubForm(formNode).get();

				assertThat(subform.getLabel()).isEqualTo(StructureMockFactory.ITEM_TITLE);
			}
		}
	}

	@Nested
	@DisplayName("Map")
	class TestMap {

		private final FormData formData = FormDataTestFactory.create();

		@Test
		void shouldSetHeader() {
			FormCycleFormHeader formHeader = mock(FormCycleFormHeader.class);
			doReturn(formHeader).when(mapper).buildHeader(formData);

			var ozgFormData = mapper.map(formData);

			assertThat(ozgFormData.getHeader()).isEqualTo(formHeader);
		}

		@Test
		void shouldSetSeviceKonto() {
			FormCycleServiceKonto serviceKonto = mock(FormCycleServiceKonto.class);
			doReturn(serviceKonto).when(mapper).buildServiceKonto(formData);

			var ozgFormData = mapper.map(formData);

			assertThat(ozgFormData.getServiceKonto()).isEqualTo(serviceKonto);
		}

		@Test
		@DisplayName("should not set service konto if postkorbId not set")
		void shouldNotSetServiceKonto() {
			var formDate = FormDataTestFactory.createBuilder().postkorbId(Optional.empty()).build();

			var ozgFormData = mapper.map(formDate);

			assertThat(ozgFormData.getServiceKonto()).isEqualTo(FormCycleServiceKonto.getDefaultInstance());
		}

		@Test
		void shouldSetAttachments() {
			FormCycleAttachmentGroup attachmentGroup = mock(FormCycleAttachmentGroup.class);
			doReturn(attachmentGroup).when(mapper).mapToAttachmentGroup(formData);

			var ozgFormData = mapper.map(formData);

			assertThat(ozgFormData.getAttachmentGroupList()).hasSize(1).first().isEqualTo(attachmentGroup);
		}

		@Test
		void shouldSetFormData() {
			GrpcFormData grpcFormData = mock(GrpcFormData.class);
			doReturn(grpcFormData).when(mapper).mapToGrpcFormData(formData);

			var ozgFormData = mapper.map(formData);

			assertThat(ozgFormData.getFormData()).isEqualTo(grpcFormData);
		}
	}

	@Nested
	@DisplayName("Map attachments")
	class TestMapAttachments {

		private final FormData formData = FormDataTestFactory.create();

		@Test
		void shouldMapAttachments() {
			var expectedAttachmentGroup = FormCycleAttachmentGroup.newBuilder().setName(OzgCloudFormDataMapper.ATTACHMENTS_GROUP_NAME)
					.addFileId(ATTACHMENT_ID).build();

			var attachmentGroup = mapper.mapToAttachmentGroup(formData);

			assertThat(attachmentGroup).isEqualTo(expectedAttachmentGroup);
		}
	}

	@Nested
	@DisplayName("Map header")
	class TestHeader {

		private final FormData formData = FormDataTestFactory.create();

		@Test
		void shouldSetReceivedTime() {
			var ozgFormDataHeader = mapper.map(formData).getHeader();

			assertThat(ozgFormDataHeader.getReceivedAt()).isNotNull();
		}

		@Test
		void shouldSetFormName() {
			var ozgFormDataHeader = mapper.map(formData).getHeader();

			assertThat(ozgFormDataHeader.getFormName()).isEqualTo(FormDataTestFactory.FORM_NAME);
		}

		@Test
		void shouldSetOrganisationsEinheitId() {
			var ozgFormDataHeader = mapper.map(formData).getHeader();

			assertThat(ozgFormDataHeader.getOrganisationsEinheitId()).isEqualTo(ORGANISATIONS_EINHEIT_ID);
		}
	}

	@Nested
	@DisplayName("Map service konto")
	class TestServiceKonto {

		private final FormData formData = FormDataTestFactory.create();

		@Test
		void shouldSetServiceKontoType() {
			var serviceKonto = mapper.buildServiceKonto(formData);

			assertThat(serviceKonto.getType()).isEqualTo(FormData.SERVICE_KONTO_TYPE);
		}

		@Test
		void shouldSetPostfachIdentifier() {
			var serviceKonto = mapper.buildServiceKonto(formData);

			assertThat(serviceKonto.getAddress().getIdentifier()).isEqualTo(POSTKORB_ID);
		}

		@Test
		void shouldSetPostfachVersion() {
			var serviceKonto = mapper.buildServiceKonto(formData);

			assertThat(serviceKonto.getAddress().getVersion()).isEqualTo(SERVICEKONTO_VERSION);
		}
	}

}
